(* Instuitionistic Multiplicative Linear Logic. *)
(**
 * This pakage contains the basis for this logic.
 * Look at {!IMLLequiv} for more.
 *)


module type FORM =
  sig
    (** The type of a formula. *)
    type t = True | Var of int | Imp of t * t | Conj of t * t

    val is_atom : t -> bool
    (** Test if the formula is an atom (ie [True] or [Var _]). *)

    val depth : t -> int
    val replace : (t -> t) -> t -> t
    val fold : ('a -> t -> 'a) -> 'a -> t -> 'a
    val max_var : t -> int
    val compare : t -> t -> int
    val string_of_form : t -> string
    val printer : Format.formatter -> t -> unit
  end

module Form : FORM =
struct

  type t =
    | True
    | Var  of int
    | Imp  of t * t
    | Conj of t * t

  let is_atom = function
    | True  -> true
    | Var _ -> true
    | _     -> false

  let rec depth =
    function
      | True -> 0
      | Var _ -> 0
      | Imp  (a, b) -> 1 + max (depth a) (depth b)
      | Conj (a, b) -> 1 + max (depth a) (depth b)

  let replace fct form =
  (* In fact it's a map function. Should it be renamed ? *)
    let rec aux = function
      | Imp  (a,b) -> fct (Imp  (aux a, aux b))
      | Conj (a,b) -> fct (Conj (aux a, aux b))
      | z -> fct z
    in aux form

  let rec fold fct acc form =
  (* TODO: verify this function is actualy used *)
    match form with
      | Imp  (a,b) ->
          let accb = fold fct acc  b in
          let acca = fold fct accb a in
          fct acca form
      | Conj (a,b) ->
          let accb = fold fct acc  b in
          let acca = fold fct accb a in
          fct acca form
      | x ->
          fct acc x

  (** @return maximum id from vars of the formula *)
  let rec max_var =
    function
      | True  -> -1
      | Var i -> i
      | Imp  (a, b) -> max (max_var a) (max_var b)
      | Conj (a, b) -> max (max_var a) (max_var b)

  let rec compare a b =
    let aux a b c d =
        let prem = compare b d
        in if prem = 0
            then compare a c
            else prem
    in
    match (a,b) with
      | (a,b) when a = b -> 0
      | (True, _) -> -1
      | (_, True) -> 1
      | (Var i, Var j) -> i - j
      | (Var _, _) -> -1
      | (_, Var _) -> 1
      | (Imp (a,b), Imp (c,d)) -> aux a b c d
      | (Imp (_,_), _) -> -1
      | (_, Imp (_,_)) -> 1
      | (Conj(a,b), Conj(c,d)) -> aux a b c d


  (** {3 printers} *)

  let rec string_of_form form =
    let fmt_binary a b op =
      let str = match (is_atom a, is_atom b) with
                  | (true,  true)  -> format_of_string   "%s %s %s"
                  | (true,  false) -> format_of_string   "%s %s (%s)"
                  | (false, true)  -> format_of_string "(%s) %s %s"
                  | (false, false) -> format_of_string "(%s) %s (%s)"
      in Printf.sprintf str (string_of_form a) op (string_of_form b)
    in
    match form with
      | True -> "I"
      | Var i when i > 0 && i < 27
              -> String.make 1 (Char.chr (96+i))
      | Var i -> string_of_int i
      | Imp  (a, b) -> fmt_binary a b "⊸"
      | Conj (a, b) -> fmt_binary a b "⊗"

  let printer ff form =
    Format.pp_print_string ff (string_of_form form)

end


(**{2 Sequents}*)

(** Data structure to hold sequents
 *
 * Input signature of {!MakeSequent}, {!MakeTheorieFromIndexed} et
 * {!MakeAllFromIndexed}.
 *)
module type INDEXED
= sig
  type 'a t

  val empty : 'a t
  val single : 'a -> 'a t
  val concat : 'a t -> 'a t -> 'a t

  val is_empty : 'a t -> bool
  val length : 'a t -> int
  val equal  : 'a t -> 'a t -> bool

  val nth : 'a t -> int -> 'a
  val find_first : ('a -> bool) -> int -> 'a t -> int

  val replace : ('a -> 'a) -> 'a t -> int -> 'a t
  val remove : 'a t -> int -> 'a t

  val map : ('a -> 'a) -> 'a t -> 'a t
  val fold_left : ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b

  val from_list : 'a list -> 'a t
  val to_list :   'a t -> 'a list
end

(** Sequents with derivation rules
 *
 *  Output signature of {!MakeSequent}.
 *  Input signature of {!MakeTheorie} and {!MakeAllFromSequent}.
 *)
module type SEQUENT
= sig

  type t

  val equal : t -> t -> bool
  val length : t -> int
  val right : t -> Form.t
  val left : t -> int -> Form.t
  val replace : (Form.t -> Form.t) -> t -> int -> t (* BEURK TODO *)
  val find_left : (Form.t -> bool) -> t -> int -> int
  val fold_left : ('a -> Form.t -> 'a) -> 'a -> t -> 'a
  val map : (Form.t -> Form.t) -> t -> t
  val depth : t -> int
  val max_var : t -> int

  val from_list : Form.t list -> t
  val make_axiom : Form.t -> t

  val is_axiom : t -> bool

  val imp_right : int -> t -> t
  val conj_left : int -> int -> t -> t
  val perm      : int -> int -> t -> t

  val cut    : t -> t -> t
  val weaken : t -> t -> t
  val imp_left : int -> t -> t -> t
  val conj_right : t -> t -> t

  val test_imp_right : t -> t -> int option
  val test_conj_left : t -> t -> (int * int) option
  val test_perm      : t -> t -> (int * int) option

  val test_imp_left : t -> t -> t -> int option

  val string_of_seq : t -> string
  val printer : Format.formatter -> t -> unit
end


(** Raised when a derivation is not possible
 *)
exception Impossible_inference


(** Make a sequent module from an indexed data structure.
 *
 * The module will have all derivation rules implemented.
 *)
module MakeSequent (S : INDEXED) : SEQUENT
= struct
  type t = Form.t S.t

  let _left seq =
    S.remove seq 0

  let _right seq =
    S.single (S.nth seq 0)

  let empty = S.empty
  let equal = S.equal
  let length = S.length
  let replace = S.replace (* BEURK TODO *)

  let right seq = S.nth seq 0
  let left seq i = S.nth seq (i + 1)
  let find_left f seq first = S.find_first f (first + 1) seq - 1
  let fold_left f acc seq = S.fold_left f acc (_left seq)
  let map = S.map

  let is_axiom seq =
    match S.length seq with
      | 0 -> true
      | 1 -> S.nth seq 0 = Form.True
      | 2 -> S.nth seq 0 = S.nth seq 1
      | _ -> false

  let depth seq =
    S.fold_left (fun d f -> max (Form.depth f) d) (-1) seq

  let max_var seq =
    S.fold_left (fun d f -> max (Form.max_var f) d) (1) seq

  let from_list = S.from_list

  let make_axiom f =
    S.from_list [f;f]



  (** {3 Derivation rules} *)

  let cut seq_a seq_b =
    match S.find_first ((=) (S.nth seq_a 0)) 1 seq_b with
      | (-1) -> raise Impossible_inference
      | i ->
          S.concat
            (_right seq_b)
            (S.concat
              (_left seq_a)
              (S.remove (_left seq_b) (i - 1))
            )

  let weaken seq_a seq_b =
    match S.nth seq_a 0 with
      | Form.True ->
          S.concat
            (_right seq_b)
            (S.concat
              (_left seq_a)
              (_left seq_b)
            )
      | _ -> raise Impossible_inference

  let conj_left i_a i_b seq =
    if i_a < 0 || i_b < 0 || i_a = i_b
      then raise Impossible_inference
      else S.replace
            (fun x -> Form.Conj (S.nth seq (i_a + 1), x))
            (S.remove seq (i_a + 1))
            (if i_a < i_b then i_b else i_b + 1)

  let conj_right seq_a seq_b =
    S.concat
      (S.single (Form.Conj (S.nth seq_a 0, S.nth seq_b 0)))
      (S.concat (_left seq_a) (_left seq_b))

  let imp_left i_b seq_a seq_b =
    S.concat
      (_right seq_b)
      (S.concat
        (_left seq_a)
        (S.replace
          (fun x -> Form.Imp (S.nth seq_a 0, x))
          (_left seq_b)
          i_b
        )
      )

  let imp_right i seq =
    if S.is_empty seq then raise Impossible_inference
    else
      S.replace
        (fun x -> Form.Imp (S.nth seq (i + 1), x))
        (S.remove seq (i + 1))
        0
  
  let perm i j seq =
    if j <= i then raise Impossible_inference else
    S.replace
      (fun _ -> left seq i)
      (S.replace (fun _ -> left seq j) seq (i + 1))
      (j + 1)

  (** {3 Derivation verification} *)

  let test_imp_right seq_up seq_down =
    match S.nth seq_down 0 with
      | Form.Imp (a,b) when b = S.nth seq_up 0 -> (
        let left_up   = _left seq_up
        and left_down = _left seq_down in
        let len_up = S.length left_up in

        if len_up != S.length left_down + 1 then None else

        let rec before i =
          let form_up = S.nth left_up i in
          if i = len_up then if form_up = a then Some i else None else
          if form_up = S.nth left_down i then before (i + 1) else
          if form_up = a then after i (i + 1)
          else None
        and after ret i =
          if i = len_up then Some ret else
          if S.nth left_up i = S.nth left_down (i - 1) then after ret (i + 1)
          else None
        in before 0 )
      | _ -> None

  let test_conj_left seq_up seq_down =
    if S.nth seq_up 0 != S.nth seq_down 0 then None else
    
    let len_up = S.length seq_up in
    if len_up != S.length seq_down + 1 then None else

    let rec pass1 i =
      if i = len_up - 1 then None else
      let form_up = S.nth seq_up i in
      match S.nth seq_down i with
        | form_down when form_down = form_up -> pass1 (i + 1)
        | Form.Conj (a, b) when b = form_up -> pass2bis a i (i + 1)
        | _ -> pass2 form_up (i - 1) (i + 1)
    and pass2 form_left ret i =
      if i = len_up then None else
      let form_up   = S.nth seq_up    i
      and form_down = S.nth seq_down (i - 1) in
      if form_up = form_down then pass2 form_left ret (i + 1) else
      if form_down = Form.Conj (form_left, form_up)
        then pass3 (ret, i - 1) (i+1)
        else None
    and pass2bis form_left ret i =
      if i = len_up - 1 then None else
      let form_up = S.nth seq_up i in
      if form_up = S.nth seq_down i then pass2bis form_left ret (i + 1) else
      if form_up = form_left then pass3 (i, ret) (i + 1)
      else None
    and pass3 ret i =
      if i = len_up then Some ret else
      if S.nth seq_up i = S.nth seq_down (i - 1) then pass3 ret (i + 1)
      else None
    in pass1 1

  let test_imp_left seq_a seq_b seq_down =
    if S.nth seq_down 0 != S.nth seq_b 0 then None else

    let left_a    = _left seq_a
    and left_b    = _left seq_b
    and left_down = _left seq_down in
    let len_a    = S.length left_a
    and len_b    = S.length left_b
    and len_down = S.length left_down in

    if len_a + len_b != len_down then None else

    let rec pass1 i =
      if i = len_a then pass2 i else
      if S.nth left_a i = S.nth left_down i then pass1 (i + 1)
      else None
    and pass2 i =
      if i = len_down then None else
      let form_b    = S.nth left_b (i - len_a) in
      match S.nth left_down i with
        | form_down when form_down = form_b ->
            pass2 (i + 1)
        | Form.Imp (a,b) when a = S.nth seq_a 0 && b = form_b ->
            pass3 (i - 1) (i + 1)
        | _ -> None
    and pass3 ret i =
      if i = len_down then Some ret else
      if S.nth left_b (i - len_a) = S.nth left_down i then pass3 ret (i + 1)
      else None
    in pass1 0

  let test_perm seq_up seq_down =
    if right seq_up != right seq_down then None else

    let len = length seq_up in
    if len != length seq_down then None else

    let rec pass1 pos =
      if pos < 0 then None else
      let form_up   = left seq_up   pos
      and form_down = left seq_down pos in
      if form_up = form_down
      then pass1 (pos-1)
      else pass2 form_down form_up pos (pos-1)

    and pass2 old_up old_down ret pos =
      if pos < 0 then None else
      let form_up   = left seq_up   pos
      and form_down = left seq_down pos in
      if form_up = form_down
      then pass2 old_up old_down ret (pos-1)
      else
        if form_up = old_up && form_down = old_down
        then pass3 (ret,pos) (pos-1)
        else None

      and pass3 ret pos =
        if pos < 0 then Some ret else
        if left seq_up pos = left seq_down pos
        then pass3 ret (pos-1)
        else None

    in pass1 (len - 2)


  (** {3 Printers} *)

  let string_of_seq seq =
    if S.is_empty seq then ""
    else
      let string_of_left left =
        if S.is_empty left then ""
        else
          S.fold_left
            (fun acc x ->
              Printf.sprintf "%s, %s"
                acc (Form.string_of_form x)
            )
            (Form.string_of_form (S.nth left 0))
            (_left left)
      in Printf.sprintf "%s → %s"
          (string_of_left (_left seq))
          (Form.string_of_form (S.nth seq 0))

  let printer ff seq =
    Format.pp_print_string ff (string_of_seq seq)

end


(**{2 Proofs}*)

module type THEORIE
= sig

  type rule_un =
      Imp_R  of int
    | Conj_L of int * int
    | Perm   of int * int

  type rule_bin =
      Cut
    | Weak
    | Imp_L of int
    | Conj_R

  (* Copied from proof.ml.
   * Not found anything better. *)
  type seq

  val seq_equal : seq -> seq -> bool

  val is_axiom : seq -> bool
  val func_un  : rule_un  -> ( seq -> seq )
  val func_bin : rule_bin -> ( seq -> seq -> seq )

  val match_un  : seq -> seq -> rule_un
  val match_bin : seq -> seq -> seq -> rule_bin

  val string_of_seq : seq -> string
  val string_of_rule_un  : rule_un  -> string
  val string_of_rule_bin : rule_bin -> string
end


module MakeTheorie (S : SEQUENT) : (THEORIE with type seq = S.t)
= struct

  type seq = S.t
  type rule_un =
    | Imp_R   of int
    | Conj_L  of int * int
    | Perm    of int * int
  type rule_bin =
    | Cut
    | Weak
    | Imp_L   of int
    | Conj_R

  let seq_equal = S.equal
  let is_axiom = S.is_axiom

  let func_un = function
    | Imp_R  i     -> S.imp_right i
    | Conj_L (i,j) -> S.conj_left i j
    | Perm   (i,j) -> S.perm      i j
  let func_bin = function
    | Cut      -> S.cut
    | Weak     -> S.weaken
    | Imp_L  i -> S.imp_left i
    | Conj_R   -> S.conj_right

  exception No_derivation of seq * (seq option) * seq

  let match_un seq_a seq_b =
    match S.test_imp_right seq_a seq_b with
      | Some v -> Imp_R v
      | None -> (
    match S.test_conj_left seq_a seq_b with
      | Some (i,j) -> Conj_L (i,j)
      | None -> (
    match S.test_perm seq_a seq_b with
      | Some (i,j) -> Perm (i,j)
      | None -> raise (No_derivation (seq_a, None, seq_b))
    ))

  let match_bin seq_a seq_b seq_c =
    match S.test_imp_left seq_a seq_b seq_c with
      | Some v -> Imp_L v
      | None ->
    let test f =
      try (S.equal (f seq_a seq_b) seq_c)
      with Impossible_inference -> false
    in
    if test S.cut         then Cut    else
    if test S.weaken      then Weak   else
    if test S.conj_right  then Conj_R else
    raise (No_derivation (seq_a, Some seq_b, seq_c))

  let string_of_seq = S.string_of_seq
  let string_of_rule_un = function
    | Imp_R   _    -> "→ ⊸"
    | Conj_L (_,_) -> "⊗ →"
    | Perm   (_,_) -> "Perm"
  let string_of_rule_bin = function
    | Cut      -> "Cut"
    | Weak     -> "Weak"
    | Imp_L  _ -> "⊸ →"
    | Conj_R   -> "→ ⊗"

end

module MakeTheorieFromIndexed (I : INDEXED)
= MakeTheorie (MakeSequent (I))

module type PROOF =
sig
  include Proof.S

  exception No_replacement of seq

  val make_axiom : Form.t -> (seq, 'a, 'b) OneTwoTree.tree
  val unit_axiom : (seq, 'a, 'b) OneTwoTree.tree
  val modus_ponens :
    Form.t -> Form.t -> optree
  val imp_trans :
    Form.t -> Form.t -> Form.t -> optree

  val cut_elimination : prooftree -> prooftree
end

module MakeProof (S : SEQUENT) (T : THEORIE with type seq = S.t)
: (PROOF with type seq       = S.t
          and type optree    = (S.t, T.rule_un, T.rule_bin) OneTwoTree.tree
          and type seqtree   = (S.t, S.t, S.t) OneTwoTree.tree
          and type prooftree = (S.t, T.rule_un, T.rule_bin) OneTwoTree.prooftree)
= struct
  include Proof.Make (T)

  exception No_replacement of S.t

  open OneTwoTree

  let make_axiom a =
    Leaf (S.make_axiom a)

  let unit_axiom = Leaf (S.from_list [Form.True])

  let modus_ponens a b =
    Binary (T.Imp_L 0,
      make_axiom a,
      make_axiom b
    )

  let imp_trans a b c =
    Unary (T.Imp_R 0,
      Binary (T.Imp_L 0,
        make_axiom a,
        modus_ponens b c
      )
    )

  let find_equiv seq_old seq_new i =
    S.find_left ((=) (S.left seq_old i)) seq_new 0

  let updt_unary_op seq_old seq_new op_old =
    let find = find_equiv seq_old seq_new in
    let find2 i j =
      let ni = find i
      and nj = find j in
      if ni = nj
      then (ni, S.find_left ((=) (S.left seq_old j)) seq_new (ni+1))
      else (ni, nj)
    in
    match op_old with
      | T.Imp_R   i    -> T.Imp_R  (find i)
      | T.Conj_L (i,j) ->
          let (ni,nj) = find2 i j
          in T.Conj_L (ni,nj)
      | T.Perm   (i,j) ->
          let (ni,nj) = find2 i j
          in T.Perm (ni,nj)

  let updt_binary_op seq_old seq_new op_old =
  (* seq_* are right seq *)
    let find = find_equiv seq_old seq_new in
    match op_old with
      | T.Imp_L i -> T.Imp_L (find i)
      | x       -> x


  let rec cut_elimination proof =
    match proof with

      | Binary ((T.Cut, _), left, right) ->
          replace_left
            (cut_elimination left)
            (cut_elimination right)

      | Leaf _ as p -> p

      | Unary ((op_old, _), branch_old) ->
          let branch_new = cut_elimination branch_old in
          let seq_branch_new = seq_from_prooftree branch_new in
          let op_new = updt_unary_op
                          (seq_from_prooftree branch_old)
                          seq_branch_new
                          op_old in
          let seq_new = T.func_un op_new seq_branch_new
          in Unary ((op_new, seq_new), branch_new)

      | Binary ((op_old, _), left_old, right_old) ->
          let right_new = cut_elimination right_old
          and left_new  = cut_elimination left_old  in
          let seq_right_new = seq_from_prooftree right_new
          and seq_left_new  = seq_from_prooftree left_new  in
          let op_new = updt_binary_op
                          (seq_from_prooftree right_old)
                          seq_right_new
                          op_old in
          let seq_new = T.func_bin op_new seq_left_new seq_right_new
          in Binary ((op_new, seq_new), left_new, right_new)
  
  and replace_left left right =
    let cutted = S.right (seq_from_prooftree left) in
    let pattern = S.make_axiom cutted in
    let rec aux = function

      | p when S.equal (seq_from_prooftree p) pattern ->
          right

      | Leaf p when S.length p = 1 && S.right p = Form.True ->
          replace_unit right

      | Unary ((T.Imp_R _, _), _) as p ->
          replace_imp p right

      | Binary ((T.Conj_R, _), _, _) as p ->
          replace_conj p right

      | Leaf s -> raise (No_replacement s)
      | Binary ((T.Cut, s), _, _) -> raise (No_replacement s)

      | Unary ((op_old, _), branch_old) ->
          let branch_new = aux branch_old in
          let seq_branch_new = seq_from_prooftree branch_new in
          let op_new = updt_unary_op
                          (seq_from_prooftree branch_old)
                          seq_branch_new
                          op_old in
          let seq_new = T.func_un op_new seq_branch_new
          in Unary ((op_new, seq_new), branch_new)

      | Binary ((op_old, _), branch_left, branch_right_old) ->
          let branch_right_new = aux branch_right_old in
          let seq_branch_right_new = seq_from_prooftree branch_right_new in
          let op_new = updt_binary_op
                          (seq_from_prooftree branch_right_old)
                          seq_branch_right_new
                          op_old in
          let seq_new = T.func_bin op_new
                          (seq_from_prooftree branch_left)
                          seq_branch_right_new
          in Binary ((op_new, seq_new), branch_left, branch_right_new)

    in aux left


  and replace_conj
    (Binary ((T.Conj_R, seq_left), left_left, left_right) as left)
    right
  =
    let cutted = S.right seq_left in
    let pattern = S.make_axiom cutted in
    let rec aux r =
      match r with

        | p when S.equal (seq_from_prooftree p) pattern ->
            left
                    
        | Unary ((T.Conj_L (i,j), seq), branch)
          when S.left seq (if i < j then j - 1 else j) = cutted ->
            replace_left
              left_left
              (replace_left left_right branch)

        | p -> right_recursion aux cutted p

    in aux right

  and replace_imp 
    (Unary ((T.Imp_R i_left, seq_left), left_left) as left)
    right
  =
    let cutted = S.right seq_left in
    let pattern = S.make_axiom cutted
    and test l r i = 
      cutted = Form.Imp (
                S.right (seq_from_prooftree l),
                S.left (seq_from_prooftree r) i  ) in
    let rec aux = function

      | p when S.equal (seq_from_prooftree p) pattern ->
          left
                  
      | Binary ((T.Imp_L i, seq), right_left, right_right)
        when test right_left right_right i ->
          replace_left
            right_left
            (replace_left left_left right_right)

      | p -> right_recursion aux cutted p

    in aux right

  and replace_unit right =
    let cutted  = Form.True in
    let pattern = Leaf (S.make_axiom cutted) in
    let rec aux = function

      | p when p = pattern -> Leaf (S.from_list [cutted])

      | p -> right_recursion aux cutted p

    in aux right

  and right_recursion fct cutted = function

      | Leaf s -> raise (No_replacement s)
      | Binary ((T.Cut, s), _, _) -> raise (No_replacement s)

      | Unary ((op_old, _), branch_old) ->
          let branch_new = fct branch_old in
          let seq_branch_new = seq_from_prooftree branch_new in
          let op_new = updt_unary_op
                          (seq_from_prooftree branch_old)
                          seq_branch_new
                          op_old in
          let seq_new = T.func_un op_new seq_branch_new
          in Unary ((op_new, seq_new), branch_new)


      | Binary ((op_old, _), left_old, right_old) ->
          let seq_left_old = seq_from_prooftree left_old in
          if S.find_left ((=) cutted) seq_left_old 0 >= 0

          then (* recursion on left *)
            let left_new = fct left_old in
            let seq_left_new = seq_from_prooftree left_new in
            let seq_new = T.func_bin op_old
                            seq_left_new
                            (seq_from_prooftree right_old)
            in Binary ((op_old, seq_new), left_new, right_old)

          else (* recursion on right *)
            let right_new = fct right_old in
            let seq_right_new = seq_from_prooftree right_new in
            let op_new = updt_binary_op
                            (seq_from_prooftree right_old)
                            seq_right_new
                            op_old in
            let seq_new = T.func_bin op_new
                            (seq_from_prooftree left_old)
                            seq_right_new
            in Binary ((op_new, seq_new), left_old, right_new)

end


(**{2 Bundles}*)

module MakeAllFromSequent (S : SEQUENT)
= struct

  module Form = Form

  module Seq = S

  module Theo = MakeTheorie (Seq)

  module OTT = OneTwoTree

  module Proof = MakeProof (Seq) (Theo)

end

module MakeAllFromIndexed (I : INDEXED)
= MakeAllFromSequent (MakeSequent (I))
