(** Instuitionistic Multiplicative Linear Logic. *)

(**{2 Logical Formulas}*)

module type FORM =
  sig
    type t = True | Var of int | Imp of t * t | Conj of t * t
    val is_atom : t -> bool
    val depth : t -> int
    val replace : (t -> t) -> t -> t
    val fold : ('a -> t -> 'a) -> 'a -> t -> 'a
    val max_var : t -> int
    val compare : t -> t -> int

    (**{3 Printers}*)

    val string_of_form : t -> string
    val printer : Format.formatter -> t -> unit
  end
module Form : FORM

(**{2 Sequents}*)

module type INDEXED =
  sig
    type 'a t
    val empty : 'a t
    val single : 'a -> 'a t
    val concat : 'a t -> 'a t -> 'a t
    val is_empty : 'a t -> bool
    val length : 'a t -> int
    val equal : 'a t -> 'a t -> bool
    val nth : 'a t -> int -> 'a
    val find_first : ('a -> bool) -> int -> 'a t -> int
    val replace : ('a -> 'a) -> 'a t -> int -> 'a t
    val remove : 'a t -> int -> 'a t
    val map : ('a -> 'a) -> 'a t -> 'a t
    val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b t -> 'a

    (**{3 Converters}*)

    val from_list : 'a list -> 'a t
    val to_list : 'a t -> 'a list
  end

module type SEQUENT =
  sig
    type t
    val equal : t -> t -> bool
    val length : t -> int
    val right : t -> Form.t
    val left : t -> int -> Form.t
    val replace : (Form.t -> Form.t) -> t -> int -> t
    val find_left : (Form.t -> bool) -> t -> int -> int
    val fold_left : ('a -> Form.t -> 'a) -> 'a -> t -> 'a
    val map : (Form.t -> Form.t) -> t -> t
    val depth : t -> int
    val max_var : t -> int
    val from_list : Form.t list -> t
    val make_axiom : Form.t -> t
    val is_axiom : t -> bool
    
    (**{3 Inferences Rules}*)

    val imp_right : int -> t -> t
    val conj_left : int -> int -> t -> t
    val perm : int -> int -> t -> t
    val cut : t -> t -> t
    val weaken : t -> t -> t
    val imp_left : int -> t -> t -> t
    val conj_right : t -> t -> t
    val test_imp_right : t -> t -> int option
    val test_conj_left : t -> t -> (int * int) option
    val test_perm : t -> t -> (int * int) option
    val test_imp_left : t -> t -> t -> int option

    (**{3 Printers}*)

    val string_of_seq : t -> string
    val printer : Format.formatter -> t -> unit
  end
exception Impossible_inference
module MakeSequent : functor (S : INDEXED) -> SEQUENT

(**{2 Proofs}*)

module type THEORIE =
  sig
    type rule_un = Imp_R of int | Conj_L of int * int | Perm of int * int
    type rule_bin = Cut | Weak | Imp_L of int | Conj_R
    type seq
    val seq_equal : seq -> seq -> bool
    val is_axiom : seq -> bool
    val func_un : rule_un -> seq -> seq
    val func_bin : rule_bin -> seq -> seq -> seq
    val match_un : seq -> seq -> rule_un
    val match_bin : seq -> seq -> seq -> rule_bin

    (**{3 Printers}*)

    val string_of_seq : seq -> string
    val string_of_rule_un : rule_un -> string
    val string_of_rule_bin : rule_bin -> string
  end
module MakeTheorie :
  functor (S : SEQUENT) -> (THEORIE with type seq = S.t)
module MakeTheorieFromIndexed :
  functor (I : INDEXED) -> THEORIE
module type PROOF =
  sig
    include Proof.S
    exception No_replacement of seq
    val make_axiom : Form.t -> (seq, 'a, 'b) OneTwoTree.tree
    val unit_axiom : (seq, 'a, 'b) OneTwoTree.tree
    val modus_ponens : Form.t -> Form.t -> optree
    val imp_trans : Form.t -> Form.t -> Form.t -> optree

    (**{3 Cut Elimination}*)

    val cut_elimination : prooftree -> prooftree
  end
module MakeProof :
  functor (S : SEQUENT) ->
  functor (T : THEORIE with type seq = S.t) ->
  (PROOF with type seq       = S.t
          and type optree    = (S.t, T.rule_un, T.rule_bin) OneTwoTree.tree
          and type seqtree   = (S.t, S.t, S.t) OneTwoTree.tree
          and type prooftree = (S.t, T.rule_un, T.rule_bin) OneTwoTree.prooftree)

(**{2 Bundles}*)

module MakeAllFromSequent :
  functor (S : SEQUENT) ->
    sig
      module Form : FORM
      module Seq : SEQUENT with type t = S.t
      module Theo : THEORIE with type seq = S.t
      module OTT :
        sig
          type ('a, 'b, 'c) tree =
            ('a, 'b, 'c) OneTwoTree.tree =
              Leaf of 'a
            | Unary of 'b * ('a, 'b, 'c) tree
            | Binary of 'c * ('a, 'b, 'c) tree * ('a, 'b, 'c) tree
          val fold_down :
            ('a -> 'b) ->
            ('b -> 'c -> 'b) ->
            ('b -> 'b -> 'd -> 'b) -> ('a, 'c, 'd) tree -> 'b
          type ('a, 'b, 'c) prooftree = ('a, 'b * 'a, 'c * 'a) tree
          val seq_from_prooftree : ('a, 'b * 'a, 'c * 'a) tree -> 'a
          val pseudo_map_down :
            ('a -> 'b -> 'c * 'a) ->
            ('a -> 'a -> 'd -> 'e * 'a) ->
            ('a, 'b, 'd) tree -> ('a, 'c * 'a, 'e * 'a) tree
          val map_down :
            ('a -> 'b) ->
            ('c -> 'd) ->
            ('e -> 'f) -> ('a, 'c, 'e) tree -> ('b, 'd, 'f) tree
          val replace :
            ('a, 'b, 'c) tree ->
            ('a, 'b, 'c) tree -> ('a, 'b, 'c) tree -> ('a, 'b, 'c) tree
        end
      module Proof :
      (PROOF with type seq       = S.t
              and type optree    = (S.t, Theo.rule_un, Theo.rule_bin) OneTwoTree.tree
              and type seqtree   = (S.t, S.t, S.t) OneTwoTree.tree
              and type prooftree = (S.t, Theo.rule_un, Theo.rule_bin) OneTwoTree.prooftree)
    end
module MakeAllFromIndexed :
  functor (I : INDEXED) ->
    sig
      module Form : FORM
      module Seq : SEQUENT
      module Theo : THEORIE with type seq = Seq.t
      module OTT :
        sig
          type ('a, 'b, 'c) tree =
            ('a, 'b, 'c) OneTwoTree.tree =
              Leaf of 'a
            | Unary of 'b * ('a, 'b, 'c) tree
            | Binary of 'c * ('a, 'b, 'c) tree * ('a, 'b, 'c) tree
          val fold_down :
            ('a -> 'b) ->
            ('b -> 'c -> 'b) ->
            ('b -> 'b -> 'd -> 'b) -> ('a, 'c, 'd) tree -> 'b
          type ('a, 'b, 'c) prooftree = ('a, 'b * 'a, 'c * 'a) tree
          val seq_from_prooftree : ('a, 'b * 'a, 'c * 'a) tree -> 'a
          val pseudo_map_down :
            ('a -> 'b -> 'c * 'a) ->
            ('a -> 'a -> 'd -> 'e * 'a) ->
            ('a, 'b, 'd) tree -> ('a, 'c * 'a, 'e * 'a) tree
          val map_down :
            ('a -> 'b) ->
            ('c -> 'd) ->
            ('e -> 'f) -> ('a, 'c, 'e) tree -> ('b, 'd, 'f) tree
          val replace :
            ('a, 'b, 'c) tree ->
            ('a, 'b, 'c) tree -> ('a, 'b, 'c) tree -> ('a, 'b, 'c) tree
        end
      module Proof :
      (PROOF with type seq       = Seq.t
              and type optree    = (Seq.t, Theo.rule_un, Theo.rule_bin) OneTwoTree.tree
              and type seqtree   = (Seq.t, Seq.t, Seq.t) OneTwoTree.tree
              and type prooftree = (Seq.t, Theo.rule_un, Theo.rule_bin) OneTwoTree.prooftree)
    end
