(* IMLLequiv *)
(*-----------*)
(** Additional operations for IMLL.
 * Those functions are used to verify commutativity of free SMCC diagrams. *)

open More

module Form
= struct
  include IMLL.Form

  let substitution_by_unit set =
    let aux = function
      | Var i when IntSet.mem i set -> True
      | x -> x
    in replace aux

  let vars form =
    let aux acc = function
      | Var i -> IntSet.add i acc
      | _ -> acc
    in fold aux (IntSet.empty) form

  let rec var_equivalence a b =
    let collapse la ra lb rb =
      match (var_equivalence la lb, var_equivalence ra rb) with
        | (Some ml, Some mr) ->
            int_inject_union ml mr
        | _ -> None
    in match (a,b) with
    | (True, True) -> Some []
      | (Var a, Var b) -> Some [(a,b)]
      | (Imp  (la,ra), Imp  (lb, rb)) ->
          collapse la ra lb rb
      | (Conj (la,ra), Conj (lb, rb)) ->
          collapse la ra lb rb
      | _ -> None

end

module FormPairSet
= struct

  module FormPair
  = struct
    type t = Form.t * Form.t

    let compare (al, ar) (bl, br) =
      let cmp_l = Form.compare al bl in
      if cmp_l != 0 then cmp_l else
      Form.compare ar br
  end

  include Set.Make (FormPair)

end

module FormMap
= struct
  include Map.Make (Form)

  let union a b =
    let f key op_a op_b =
      match (op_a, op_b) with
        | (Some v, None) -> Some v
        | (None, Some v) -> Some v
        | (Some va, Some vb) when va = vb -> Some va
        | (None, None) -> None
        | _ -> failwith "FormMap.union impossible"
    in merge f a b

end

module ImpPartSet
= struct

  module ImpPart
  = struct
    type t = IntSet.t * Form.t * IntSet.t

    let compare (set_a, form_a, _) (set_b, form_b, _) =
      if IntSet.subset set_a set_b then -1 else
      if IntSet.subset set_b set_a then  1 else
      Form.compare form_a form_b
  end

  include Set.Make (ImpPart)
end

module ExtendSequent (S : IMLL.SEQUENT)
= struct
  open Form
  include S

  let substitution_by_unit set seq =
    map (substitution_by_unit set) seq

  let vars seq =
    fold_left
      (fun acc v -> IntSet.union (Form.vars v) acc)
      (Form.vars (right seq))
      seq

  let aux_parts old pos seq =
    let vars = Form.vars (left seq pos) in
    let rec aux acc (pos,vars) = function
      | [] -> (pos,vars)::acc
      | (o_pos, o_vars)::q when IntSet.has_inter vars o_vars ->
          aux acc (IntSet.union pos o_pos, IntSet.union vars o_vars) q
      | t::q ->
          aux (t::acc) (pos, vars) q
    in aux [] (IntSet.singleton pos, vars) old

  let parts seq =
    let rec aux acc pos =
      if pos < 0
        then acc
        else aux (aux_parts acc pos seq) (pos - 1)
    in aux [] (length seq - 2)

  (** 
   * Collapse left part of the sequent (IntSet.t * IntSet.t) list
   * Scan list : if not in same -> None else Some collapsed perm
   *)
  let filter_condition_2 seq (a, b) =
    let collapsed = parts seq in
    let rec scan acc = function
      | [] -> Some acc
      | (pos, _)::q when IntSet.mem a pos ->
          if IntSet.mem b pos
            then scan acc q
            else None
      | (pos, _)::q when IntSet.mem b pos -> None
      | (_, var)::q ->
          scan (IntSet.union acc var) q
    in scan IntSet.empty collapsed

  let imp_parts seq =
    let rec aux acc pos =
      if pos < 0 then acc else
      match left seq pos with
        | Imp (a,b) ->
            let nseq = replace (fun _ -> a) seq (pos + 1) in (* BEURK TODO *)
            let nseq_parts = parts nseq in
            if List.length nseq_parts < 2 then aux acc (pos-1) else
            let (poss, vars) =
              List.find
                (fun (p,_) -> IntSet.mem pos p)
                nseq_parts
              in
            if IntSet.has_inter vars (Form.vars b)
              then aux acc (pos-1)
              else aux (ImpPartSet.add (poss, a, vars) acc) (pos-1)
        | _ -> aux acc (pos-1)
    in aux ImpPartSet.empty (length seq - 2)

  let var_equivalence_aux seq len_seq mapping form =
    let all_mappings_for_form =
      let rec add_mapping acc i =
        if i < 0 then acc else
        match Form.var_equivalence form (S.left seq i) with
          | Some m -> add_mapping ((IntSet.singleton i, m)::acc) (i-1)
          | None   -> add_mapping acc (i-1)
      in add_mapping [] (len_seq-1) in
    let all_mappings = List.cartesian_product mapping all_mappings_for_form in
    let rec collapse acc = function
      | [] -> acc
      | ((set_a, map_a), (set_b, map_b))::q ->
          ( match ( IntSet.has_inter set_a set_b,
                    int_inject_union map_a map_b ) with
              | (false, Some map) ->
                    collapse ( (IntSet.union set_a set_b, map)::acc ) q
              | _ ->
                    collapse acc q
          )
    in collapse [] all_mappings
      
  let var_equivalence_perm seq_a seq_b =
    match Form.var_equivalence (right seq_a) (right seq_b) with
      | Some m ->
          let lst = S.fold_left
                      (var_equivalence_aux seq_b (length seq_b - 1))
                      [(IntSet.empty, m)]
                      seq_a
          in ( match lst with
                | (_, vars_map)::_ -> Some vars_map
                | [] -> None )
      | None -> None


  let var_equivalence_strict seq_a seq_b =
    let rec aux ret pos =
      if pos < 0 then ret else
      match ret with
        | None -> None
        | Some ident ->
            ( match Form.var_equivalence (left seq_a pos) (left seq_b pos) with
                | None -> None
                | Some nv_ident ->
                    aux (int_inject_union ident nv_ident) (pos - 1)
            )
      in 
    let len_a = length seq_a
    and len_b = length seq_b in
    if len_a = len_b
    then
      aux (Form.var_equivalence (right seq_a) (right seq_b)) (len_a - 2)
    else
      None

  let contains_double_implication form_a form_b seq =
    let get_var = function
      | Imp (Imp(a,b), _) -> (a,b)
      | _ -> failwith "contains_double_implication: not a double implication" in
    let (fa,sa) = get_var form_a
    and (fb,sb) = get_var form_b in
    let search_one (f,s) ret pos =
      let rec aux pos =
        if pos < 0 then None else
        match S.left seq pos with
          | Imp (Imp(a,b), True) when a = f && b = s ->
              Some (ret, pos)
          | Imp (Imp(a,True), True) when a = f ->
              Some (ret, pos)
          | _ -> aux (pos-1)
      in aux pos in
    let rec search_two pos =
      if pos < 0 then None else
      match S.left seq pos with
        | Imp (Imp(a,b), True) when a = fa && b = sa ->
            search_one (fb, sb) pos (pos-1)
        | Imp (Imp(a,b), True) when a = fb && b = sa ->
            search_one (fa, sa) pos (pos-1)
        | Imp (Imp(a,True), True) when a = fa ->
            search_one (fb, sb) pos (pos-1)
        | Imp (Imp(a,True), True) when a = fb ->
            search_one (fa, sa) pos (pos-1)
        | _ -> search_two (pos-1)
    in search_two (length seq - 2)

  let substitution_conj_var seq =
  (** return substitution for all formula of the form a or a x b.
   * seq MUST be a pure 2-seq *)
    let f set = function
      | Var i -> IntSet.add i set
      | Conj (Var i, Var j) ->
          IntSet.add i (IntSet.add j set)
      | _ -> set
    in fold_left f IntSet.empty seq

end

module type EXTENDED_SEQUENT
= sig
  include IMLL.SEQUENT

  val substitution_by_unit : IntSet.t -> t -> t
  val vars : t -> IntSet.t
  val imp_parts : t -> ImpPartSet.t
  val filter_condition_2 : t -> int * int -> IntSet.t option
  val contains_double_implication : Form.t -> Form.t -> t -> (int * int) option
  val substitution_conj_var : t -> IntSet.t

end

module MakeProof (S : EXTENDED_SEQUENT)
                 (T : IMLL.THEORIE with type seq = S.t)
= struct
  include IMLL.MakeProof (S) (T)
  open OneTwoTree
  open IMLL.Form

  let aux221 a b c d i =
    Unary (T.Imp_R 0,
      Binary (T.Imp_L i,
        make_axiom a,
        imp_trans b c d
      )
    )

  let aux222 a b c i =
  Unary (T.Imp_R 0,
      Binary (T.Imp_L i,
        make_axiom a,
        Unary (T.Conj_L (0, 1 + i),
          Binary (T.Conj_R,
            b,
            c
          )
        )
      )
    )

  let aux231 a b c d i =
    Unary (T.Imp_R i,
      Binary (T.Imp_L 0,
        imp_trans a b c,
        make_axiom d
      )
    )

  let aux232 a b c i =
    Unary (T.Imp_R i,
      Binary (T.Imp_L 0,
        Unary (T.Conj_L (0, 1 + i),
          Binary (T.Conj_R,
            a,
            b
          )
        ),
        make_axiom c
      )
    )

  let depth_red_form var form =
    let p = Var var in
    match form with                         (* Case from algorithm *)

      | Conj (a, b) ->                      (* 1 *)
          Binary (T.Conj_R,
            make_axiom a,
            make_axiom b
          )

      | Imp (a, Imp (b, c))
        when is_atom a && is_atom b ->      (* 2.2.1.2 *)
          aux221 a b p c 0

      | Imp (a, Imp (b, c))
        when is_atom a ->                   (* 2.2.1.1 *)
          aux221 a b p c 1

      | Imp (a, Conj (b, c))
        when is_atom a && is_atom b ->      (* 2.2.2.2 *)
          aux222 a (make_axiom b) (modus_ponens p c) 0

      | Imp (a, Conj (b, c))
        when is_atom a ->                   (* 2.2.2.1 *)
          aux222 a (modus_ponens p b) (make_axiom c) 1

      | Imp (Imp (a, b), c)
        when is_atom a && is_atom c ->      (* 2.3.1.2 *)
          aux231 a b p c 0

      | Imp (Imp (a, b), c)
        when is_atom c ->                   (* 2.3.1.1 *)
          aux231 p a b c 1

      | Imp (Conj (a, b), c)
        when is_atom a && is_atom c ->      (* 2.3.2.2 *)
          aux232 (make_axiom a) (modus_ponens b p) c 0

      | Imp (Conj (a, b), c)
        when is_atom c ->                   (* 2.3.2.1 *)
          aux232 (modus_ponens a p) (make_axiom b) c 1

      | Imp (a, b) ->                       (* 2.1 *)
          imp_trans a p b

      | _ -> failwith "Depth was not greater than 2"

  let depth_red_gen f seq =
    let pos = S.find_left (fun f -> IMLL.Form.depth f > 2) seq 0 in
    if pos < 0 then None else
    let form = S.left seq pos in
    Some (f (S.max_var seq + 1) form)

  let rec aux_depth_red proof =
  (* Implements recursively the depth reduction algorithm to a proof,
   * returning an annotated proof tree. *)
    let mk_proof v f = fromOp (depth_red_form v f)
    and seq = seq_from_prooftree proof in
    match depth_red_gen mk_proof seq with
      | None -> proof
      | Some s ->
          let cut_seq = S.cut (seq_from_prooftree s) seq
          in aux_depth_red (Binary ((T.Cut, cut_seq),
                          s,
                          proof
                        ))

  let depth_red proof =
    let seq = seq_from_prooftree proof in
    let imp_right = make_axiom (Var (S.max_var seq + 1)) in
    let new_seq = S.imp_left 0 seq (seq_from_prooftree imp_right) in
    aux_depth_red (Binary ((T.Imp_L 0, new_seq), proof, imp_right ))

  let aux_purification acc form =
    match form with
      | Imp (a, Imp(b, c)) ->
              (Unary (T.Imp_R 0,
                Unary (T.Imp_R 1,
                  Binary (T.Imp_L 0,
                    Binary (T.Conj_R,
                      make_axiom a,
                      make_axiom b
                    ),
                    make_axiom c
                  )
                )
              ))::acc

      | Conj (a, b) when Form.depth form > 1 ->
              (Binary (T.Conj_R,
                make_axiom a,
                make_axiom b
              ))::acc

      | _ -> acc
                    
  let purification proof =
    let cuts = S.fold_left aux_purification [] (seq_from_prooftree proof) in
    let aux_cut proof subproof =
      let cut_proof = fromOp subproof in
      let new_seq = S.cut (seq_from_prooftree cut_proof)
                          (seq_from_prooftree proof)
      in Binary ((T.Cut, new_seq), cut_proof, proof)
    in List.fold_left aux_cut proof cuts

  let aux_imp a =
  (* a -> I -o A *)
    Unary (T.Imp_R 0,
      Binary (T.Weak,
        make_axiom True,
        make_axiom a
      )
    )

  let aux_conj1 a swap =
  (* A -> I x A
    * A -> A x I *)
    let (left, right) = swapif swap unit_axiom (make_axiom a) in
    Binary (T.Conj_R, left, right)

  let aux_conj2 a b swap_up swap_down =
    let (left, right) = swapif swap_down
                          (aux_conj1 a swap_up)
                          (make_axiom b)
    in  Unary (T.Conj_L (0,1),
          Binary (T.Conj_R, left, right)
        )

  let aux_mix_imp1 a b swap =
    let (i,j) = swapif swap 0 1 in
    Unary (T.Imp_R 0,
      Unary (T.Conj_L (i,j),
        Binary (T.Weak,
          make_axiom True,
          modus_ponens a b
        )
      )
    )

  let aux_mix_imp2 a b swap =
    let (left, right) = swapif swap unit_axiom (modus_ponens a b) in
    Unary (T.Imp_R 0,
      Binary (T.Conj_R, left, right)
    )

  let aux_mix_conj a b swap =
    let (left, right) = swapif swap (make_axiom a) (aux_imp b) in
    Unary (T.Conj_L (0,1),
      Binary (T.Conj_R, left, right)
    )

  let rec aux_simpl_i seq i =
    if i < 0 then None else
    match S.left seq i with

      | True ->
          Some( unit_axiom )

      | Imp (True, a) ->
          Some( aux_imp a )

      | Conj (True, a) ->
          Some( aux_conj1 a false )

      | Conj (a, True) ->
          Some( aux_conj1 a true )

      | Imp (Imp (True, a), b) ->
          Some( Unary (T.Imp_R 0,
                  Binary (T.Imp_L 0,
                    unit_axiom,
                    modus_ponens a b
                  )
                )
              )

      | Conj (Conj (True, a), b) ->
          Some( aux_conj2 a b false false )

      | Conj (Conj (a, True), b) ->
          Some( aux_conj2 a b true false )

      | Imp (a, Imp(True, b)) ->
          Some( Unary (T.Imp_R 0,
                  Binary (T.Imp_L 0,
                    make_axiom a,
                    aux_imp b
                  )
                )
              )

      | Conj (a, Conj(True, b)) ->
          Some( aux_conj2 b a false true )

      | Conj (a, Conj(b, True)) ->
          Some( aux_conj2 b a true true )

      | Imp (Conj (True, a), b) ->
          Some( aux_mix_imp1 a b false )

      | Imp (Conj (a, True), b) ->
          Some( aux_mix_imp1 a b true )

      | Imp (a, Conj(True, b)) ->
          Some( aux_mix_imp2 a b false )

      | Imp (a, Conj(b, True)) ->
          Some( aux_mix_imp2 a b true )

      | Conj (a, Imp(True, b)) ->
          Some( aux_mix_conj a b false )

      | Conj (Imp(True, a), b) ->
          Some( aux_mix_conj b a true )

      | _ -> aux_simpl_i seq (i - 1)

  let rec simplify_left proof =
    let seq = seq_from_prooftree proof in
    match aux_simpl_i seq (S.length seq - 2) with
      | None -> proof
      | Some p_op -> 
        let p = fromOp p_op in
        let cut_seq = S.cut (seq_from_prooftree p) seq
        in simplify_left (Binary ((T.Cut, cut_seq),
                            p,
                            proof
                         ))

  let aux_right_conj form on_left =
    let (i,j) = swapif on_left 1 0 in
    Unary (T.Conj_L (i,j),
      Binary (T.Weak,
        make_axiom True,
        make_axiom form
      )
    )

  let simplify_right proof =
    let seq = seq_from_prooftree proof in
    let mk_proof cut_op =
      let cut_proof = fromOp cut_op in
      let cut_seq   = S.cut seq (seq_from_prooftree cut_proof) in
      Binary ((T.Cut, cut_seq), proof, cut_proof)
    in
    match S.right seq with

      | Imp (True, a) ->
          mk_proof (Binary (T.Imp_L 0, unit_axiom, make_axiom a))

      | Conj (True, a) ->
          mk_proof (aux_right_conj a true)

      | Conj (a, True) ->
          mk_proof (aux_right_conj a false)

      | _ -> proof

  let simplify proof =
    simplify_left (simplify_right proof)

  let rec atomisation proof =
    let aux seq =
      match (S.is_axiom seq, S.right seq) with
        | (false, _) -> failwith "Leaf is not an axiom"

        | (_, Imp (a,b)) ->
            atomisation
              (fromOp
                (Unary (T.Imp_R 0,
                  modus_ponens a b
                ))
              )

        | (_, Conj (a,b)) ->
            atomisation
              (fromOp
                (Unary (T.Conj_L (0,1),
                  Binary (T.Conj_R,
                    make_axiom a,
                    make_axiom b
                  )
                ))
              )

        | _ -> Leaf seq

    in fold_down
      aux
      (fun a v -> Unary (v,a))
      (fun a b v -> Binary (v,a,b))
      proof


  let normalize_var proof =
  (** Make a balanced proof from any prooftree *)
    let next_var = ref 1 in
    let replace_var vars =
      let rec f x i = function
        | [] -> raise Not_found
        | t::q when t = x -> i
        | _::q -> f x (i+1) q
      in S.map
        (Form.replace
          (function Var x -> Var (f x !next_var vars) | x -> x) ) in
    let f seq =
      let right = S.right seq in
      if S.is_axiom seq && right != Form.True
        then
          let vars = IntSet.elements (Form.vars right) in
          let ret = replace_var vars seq in
          (next_var := !next_var + List.length vars ; ret)
        else seq
    in fromOp (OneTwoTree.map_down f fst fst proof)
    
  let vars_identification vars_map proof =
  (** proof MUST be atomic *)
    let compressed_map =
      List.fold_left
        (fun acc (a,b) -> if a = b then acc else (a,b)::acc)
        []
        vars_map
      in
    let map_leaf seq =
      match S.right seq with
        | Var i ->
            ( try S.make_axiom (Var (List.assoc i compressed_map))
              with Not_found -> seq )
        | _     -> seq
    in fromOp (OneTwoTree.map_down map_leaf fst fst proof)

  let find_KF_candidate a b =
  (** find KF-candidate from a and b (allready var equals)
   * with respect to ordering of -o -> in proofs.
   *)
    let find_in_proof swap proof =
      let rec aux bigger = function
        | Leaf _ -> (FormMap.empty, FormPairSet.empty)
        | Unary  (_, p) -> aux bigger p
        | Binary ((op, seq), pa, pb) ->
            ( match (op, S.right (seq_from_prooftree pa)) with
                | (T.Imp_L pos, (Imp (_,_) as form)) ->
                    let seq_right = seq_from_prooftree pb in
                    let imp_form = Imp (form, S.left seq_right pos) in
                    let (map, set) = uni (imp_form::bigger) pa pb
                    in (FormMap.add imp_form (S.vars seq_right) map,
                        List.fold_left
                          (fun acc v ->
                            FormPairSet.add (swapif swap v imp_form) acc)
                          set
                          bigger)
                | _ -> uni bigger pa pb
            )
      and uni bigger pa pb =
        let (map_a, set_a) = aux bigger pa
        and (map_b, set_b) = aux bigger pb
        in (FormMap.union     map_a map_b,
            FormPairSet.union set_a set_b)
      in aux [] proof in
    let (map_a, kf_a) = find_in_proof false a
    and (map_b, kf_b) = find_in_proof true  b
    in FormPairSet.fold
        (fun (form_a, form_b) acc ->
          ( form_a, form_b,
            IntSet.union  (FormMap.find form_a map_a)
                          (FormMap.find form_b map_b)
          )::acc)
        (FormPairSet.inter kf_a kf_b)
        []

  let substitution_by_unit set =
    let aux (a,seq) = (a, S.substitution_by_unit set seq) in
    map_down (S.substitution_by_unit set) aux aux

  let star_operation subst proof =
    List.fold_left
      (fun p f -> f p)
      proof
      [substitution_by_unit subst; simplify; cut_elimination]

  let is_equiv_term_imp_l p =
    let seq = seq_from_prooftree p in
    let all_vars = S.vars seq in
    let rec aux = function
      | [] -> None
      | (_,form,vars)::q ->
          ( match star_operation (IntSet.diff all_vars vars) p with
              | Binary ((T.Imp_L _,_), premisse_left, _)
                when S.right (seq_from_prooftree premisse_left) = form ->
                  Some vars
              | _ -> aux q
          )
    in aux (ImpPartSet.elements (S.imp_parts seq))

  let critical_pair proof_a proof_b =
  (** proof_a and proof_b MUST be simplified, atomic, balanced.
   * Their final sequent MUST be pure 2-sequent and equivalent up to formula
   * position (same var names).
   *)
    let rec aux_cp = function
      | [] -> None
      | (fa,fb,subst)::q -> aux_cp_inner (proof_a, proof_b) (fa,fb) subst q
    and aux_cp_inner (pa_arg,pb_arg) (fa,fb) subst q =
      let pa = star_operation subst pa_arg
      and pb = star_operation subst pb_arg in
      let seq = seq_from_prooftree pa in

      let contains = S.contains_double_implication fa fb seq in
      if contains = None then aux_cp q else
      let Some (pos_fa, pos_fb) = contains in

      let cv_subst = S.substitution_conj_var seq in
      if not (IntSet.is_empty cv_subst)
        then aux_cp_inner (pa,pb) (fa,fb) cv_subst q else

      match S.filter_condition_2 seq (pos_fa, pos_fb) with
        | None -> aux_cp q
        | Some subst ->
            let ppa = star_operation subst pa
            and ppb = star_operation subst pb in
            ( match (ppa, ppb) with
                | ( Binary ((T.Imp_L _,_), left_a, _),
                    Binary ((T.Imp_L _,_), left_b, _) ) ->
                      aux_cp_cond3 (left_a, left_b) (ppa, ppb) (fa,fb) q
                | _ -> failwith "critical_pair: not an Imp_L after condition 1"
            )

    and aux_cp_cond3 (left_a, left_b) (pa, pb) (fa, fb) q =
      match is_equiv_term_imp_l left_a with
        | Some subst -> aux_cp_inner (pa,pb) (fa,fb) subst q
        | None ->
            ( match is_equiv_term_imp_l left_b with
                | Some subst -> aux_cp_inner (pa,pb) (fa,fb) subst q
                | None -> Some (fa,fb)
            )

    in aux_cp (find_KF_candidate proof_a proof_b)

end

module MakeAllFromSequent (S : IMLL.SEQUENT)
= struct
  module Form = Form

  module Seq = ExtendSequent (S)

  module Theo = IMLL.MakeTheorie (Seq)

  module OTT = OneTwoTree

  module Proof = MakeProof (Seq) (Theo)

end

module MakeAllFromIndexed (I : IMLL.INDEXED)
= MakeAllFromSequent (IMLL.MakeSequent (I))
