(** Minimalistic Implicative Logic with constants. *)

(* Logical formulas *)
module Form =
struct

type t =
    | True
    | False
    | Var of int
    | Imp of t * t

let rec simplification form =
  match form with
    | Imp (a,b) -> 
        ( match (simplification a, simplification b) with
            | (True, b)  -> b
            | (False, _) -> True
            | (_, True)  -> True
            | (a,b)      -> Imp (a,b)
        )
    | f -> f

(* printers *)

let rec string_of_form form =
  match form with
    | True -> "⊤"
    | False -> "⊥"
    | Var i when i > 0 && i < 27
            -> String.make 1 (Char.chr (96+i))
    | Var i -> string_of_int i
    | Imp ((Imp (_,_) as a), (Imp (_,_) as b))
            -> Printf.sprintf "(%s) → (%s)" (string_of_form a) (string_of_form b)
    | Imp (Imp (_,_) as a, b)
            -> Printf.sprintf "(%s) → %s" (string_of_form a) (string_of_form b)
    | Imp (a, (Imp (_,_) as b))
            -> Printf.sprintf "%s → (%s)" (string_of_form a) (string_of_form b)
    | Imp (a,b)
            -> Printf.sprintf "%s → %s" (string_of_form a) (string_of_form b)

let printer ff form =
  Format.pp_print_string ff (string_of_form form)

(* Order *)

let rec compare a b =
  match (a,b) with
    | (a,b) when a = b -> 0
    | (True, _) -> -1
    | (_, True) -> 1
    | (False, _) -> -1
    | (_, False) -> 1
    | (Var i, Var j) -> i - j
    | (Var _, _) -> -1
    | (_, Var _) -> 1
    | (Imp (a,b), Imp (c,d)) -> (
        (* On classe selon la conséquence d'abord pour faciliter
         * Verif.test_imp_g. *)
        let prem = compare b d
        in if prem = 0
            then compare a c
            else prem
    )

end

module type INDEXED
= sig
  type 'a t

  val empty : 'a t
  val single : 'a -> 'a t
  val concat : 'a t -> 'a t -> 'a t

  val is_empty : 'a t -> bool
  val length : 'a t -> int
  val equal  : 'a t -> 'a t -> bool

  val nth : 'a t -> int -> 'a
  val find_first : ('a -> bool) -> int -> 'a t -> int

  val replace : ('a -> 'a) -> 'a t -> int -> 'a t
  val remove : 'a t -> int -> 'a t

  val map : ('a -> 'a) -> 'a t -> 'a t
  val fold_left : ('b -> 'a -> 'b) -> 'b -> 'a t -> 'b
end


module MakeSequent (S : INDEXED)
= struct
  type t = Form.t S.t

  let empty = S.empty
  let equal = S.equal

  let _left seq =
    S.remove seq 0

  let _right seq =
    S.single (S.nth seq 0)

  let imp_left i_b seq_a seq_b =
    match (S.is_empty seq_a, S.is_empty seq_b) with
      | (true, _) -> seq_b
      | (_, true) -> seq_a
      | _ ->
          S.concat
            (_right seq_b)
            (S.concat
              (_left seq_a)
              (S.replace
                (fun x -> Form.Imp (S.nth seq_a 0, x))
                (_left seq_b)
                i_b
              )
            )

  let imp_right i seq =
    if S.is_empty seq then empty
    else
      S.replace
        (fun x -> Form.Imp (S.nth seq (i + 1), x))
        (S.remove seq (i + 1))
        0

  let weak_left form seq =
    let s = S.single form in
    if S.is_empty seq
    then S.concat s s
    else
      S.concat
        (_right seq)
        (S.concat s (_left seq))

  let weak_right form seq =
    let s = S.single form in
    if S.is_empty seq
    then S.concat s s
    else
      match S.nth seq 0 with
        | Form.False -> S.concat s (_left seq)
        | _ -> seq

  let contractions =
    let rec aux n seq =
      if n >= S.length seq then seq
      else
        match S.find_first ((=) (S.nth seq n)) (n+1) seq with
          | (-1) -> aux (n+1) seq
          | i -> aux n (S.remove seq i)
    in aux 1

  let cut seq_a seq_b =
    match S.find_first ((=) (S.nth seq_a 0)) 1 seq_b with
      | (-1) -> S.empty
      | i ->
          S.concat
            (_right seq_b)
            (S.concat
              (_left seq_a)
              (S.remove (_left seq_b) i)
            )

  let simplification seq =
    let f_simpl = S.map Form.simplification seq in
    if  S.nth f_simpl 0 = Form.True ||
        S.find_first ((=) Form.False) 1 f_simpl != -1
    then S.single Form.True
    else
      let rec aux s i =
        match S.find_first ((=) Form.True) i s with
          | (-1) -> s
          | i -> aux (S.remove s i) i
      in aux f_simpl 1
  
  let is_axiom seq =
    let len = S.length seq in
    len = 0 ||
    (len = 2 && S.nth seq 0 = S.nth seq 1)

  let test_imp_right seq_up seq_down =
    match S.nth seq_down 0 with
      | Form.Imp (a,b) when b = S.nth seq_up 0 -> (
        let left_up   = _left seq_up
        and left_down = _left seq_down in
        let len_up = S.length left_up in

        if len_up != S.length left_down + 1 then None else

        let rec before i =
          let form_up = S.nth left_up i in
          if i = len_up then if form_up = a then Some i else None else
          if form_up = S.nth left_down i then before (i + 1) else
          if form_up = a then after i (i + 1)
          else None
        and after ret i =
          if i = len_up then Some ret else
          if S.nth left_up i = S.nth left_down (i - 1) then after ret (i + 1)
          else None
        in before 0 )
      | _ -> None

  let test_weak_left seq_up seq_down =
    let len_down = S.length seq_down in

    if S.length seq_up + 1 != len_down then None else
    if S.nth seq_up 0 != S.nth seq_down 0 then None else

    let rec pass i =
      if i = len_down then Some (S.nth seq_down 1) else
      if S.nth seq_up (i - 1) = S.nth seq_down i then pass (i + 1)
      else None
    in pass 2

  let test_weak_right seq_up seq_down =
    match S.nth seq_up 0 with
      | Form.False -> Some (S.nth seq_down 0)
      | _ -> None

  let test_imp_left seq_a seq_b seq_down =
      if S.nth seq_down 0 != S.nth seq_b 0 then None else

      let left_a    = _left seq_a
      and left_b    = _left seq_b
      and left_down = _left seq_down in
      let len_a    = S.length left_a
      and len_b    = S.length left_b
      and len_down = S.length left_down in

      if len_a + len_b != len_down then None else

      let rec pass1 i =
        if i = len_a then pass2 i else
        if S.nth left_a i = S.nth left_down i then pass1 (i + 1)
        else None
      and pass2 i =
        if i = len_down then None else
        let form_b    = S.nth left_b (i - len_a) in
        match S.nth left_down i with
          | form_down when form_down = form_b ->
              pass2 (i + 1)
          | Form.Imp (a,b) when a = S.nth seq_a 0 && b = form_b ->
              pass3 (i - 1) (i + 1)
          | _ -> None
      and pass3 ret i =
        if i = len_down then Some ret else
        if S.nth left_b (i - len_a) = S.nth left_down i then pass3 ret (i + 1)
        else None
      in pass1 0
        

  let string_of_seq seq =
    if S.is_empty seq then "∅"
    else
      let string_of_left left =
        if S.is_empty left then ""
        else
          S.fold_left
            (fun acc x ->
              Printf.sprintf "%s, %s"
                acc (Form.string_of_form x)
            )
            (Form.string_of_form (S.nth left 0))
            (_left left)
      in Printf.sprintf "%s ⊦ %s"
          (string_of_left (_left seq))
          (Form.string_of_form (S.nth seq 0))

  let printer ff seq =
    Format.pp_print_string ff (string_of_seq seq)

end

module type SEQUENT
= sig

  type t

  val is_axiom : t -> bool
  val equal : t -> t -> bool

  val imp_right : int -> t -> t
  val weak_left : Form.t -> t -> t
  val weak_right : Form.t -> t -> t
  val contractions : t -> t
  val simplification : t -> t

  val imp_left : int -> t -> t -> t
  val cut : t -> t -> t

  val test_imp_right  : t -> t -> int option
  val test_weak_left  : t -> t -> Form.t option
  val test_weak_right : t -> t -> Form.t option

  val test_imp_left : t -> t -> t -> int option

  val string_of_seq : t -> string
end


module MakeTheorieFromSequent (S : SEQUENT)
= struct

  type seq = S.t
  type rule_un =
    | Imp_R   of int
    | Weak_L  of Form.t
    | Weak_R  of Form.t
    | Contr
    | Simpl
  type rule_bin =
    | Imp_L   of int
    | Cut

  let seq_equal = S.equal
  let is_axiom = S.is_axiom

  let func_un = function
    | Imp_R  i -> S.imp_right i
    | Weak_L f -> S.weak_left f
    | Weak_R f -> S.weak_right f
    | Contr    -> S.contractions
    | Simpl    -> S.simplification
  let func_bin = function
    | Imp_L  i -> S.imp_left i
    | Cut      -> S.cut

  exception No_derivation of seq * (seq option) * seq

  let match_un seq_a seq_b =
    match S.test_imp_right seq_a seq_b with
      | Some v -> Imp_R v
      | None -> (
    match S.test_weak_left seq_a seq_b with
      | Some v -> Weak_L v
      | None -> (
    match S.test_weak_right seq_a seq_b with
      | Some v -> Weak_R v
      | None -> 
    if S.equal (S.contractions seq_a) seq_b
      then Contr else
    if S.equal (S.simplification seq_a) seq_b
      then Simpl else
    raise (No_derivation (seq_a, None, seq_b))
      ))

  let match_bin seq_a seq_b seq_c =
    match S.test_imp_left seq_a seq_b seq_c with
      | Some v -> Imp_L v
      | None -> 
    if S.equal (S.cut seq_a seq_b) seq_c
      then Cut else
    raise (No_derivation (seq_a, Some seq_b, seq_c))

  let string_of_seq = S.string_of_seq
  let string_of_rule_un = function
    | Imp_R  _ -> "⊦ →"
    | Weak_L _ -> "Weak ⊦"
    | Weak_R _ -> "⊦ Weak"
    | Contr    -> "Contr"
    | Simpl    -> "Simpl"
  let string_of_rule_bin = function
    | Imp_L  _ -> "→ ⊦"
    | Cut      -> "Cut"

end

module MakeTheorieFromIndexed (I : INDEXED)
= MakeTheorieFromSequent (MakeSequent (I))
