(** Indexed structures using trees. *)

open More

module Unbalanced
= struct
  type 'a t =
    | Leaf
    | Node of 'a * int * 'a t * 'a t

  let empty = Leaf
  let single x = Node (x, 1, Leaf, Leaf)

  let is_empty = function
    | Leaf -> true
    | _ -> false

  let length = function
    | Leaf -> 0
    | Node (_,n,_,_) -> n

  let rec equal a b =
    let rec aux stack_a stack_b =
      match (stack_a, stack_b) with
        | ([], []) -> true
        | (Leaf::qa, sb) -> aux qa sb
        | (sa, Leaf::qb) -> aux sa qb
        | ((Node (va, 1, Leaf, Leaf))::qa, (Node (vb, 1, Leaf, Leaf))::qb)
          -> va = vb && aux qa qb
        | ((Node (va, na, la, ra))::qa, (Node (vb, nb, lb, rb))::qb)
          when na = nb && length la = length lb
          -> va = vb && equal la lb && equal ra rb && aux qa qb
        | ((Node (va, na, la, ra))::qa, (Node (vb, nb, lb, rb))::qb)
          -> aux (la :: (Node (va, 1, Leaf, Leaf)) :: ra :: qa)
                 (lb :: (Node (vb, 1, Leaf, Leaf)) :: rb :: qb)
        | _ -> false
    in aux [a] [b]

  let rec nth tree i =
    match tree with
      | Node (x, n, a, b) ->
          ( match i - length a with
              | n when n < 0 -> nth a i
              | 0 -> x
              | n when n > 0 -> nth b (n - 1)
              | _ -> failwith "No match"
          )
      | _ -> raise Not_found

  let rec remove tree i =
    match tree with
      | Node (x, sz, a, b) ->
          ( match i - length a with
              | n when n < 0 ->
                  Node (x, sz - 1, remove a i, b)
              | n when n > 0 ->
                  Node (x, sz - 1, a, remove b (n - 1))
              | 0 ->
                  concat a b
              | _ -> failwith "No match"
          )
      | _ -> raise Not_found

  and concat a b =
    match (length a, length b) with
      | (0,_) -> b
      | (_,0) -> a
      | (len_a, len_b) when len_a > len_b ->
          let root = nth a (len_a - 1)
          and a = remove a (len_a - 1)
          in Node (root, len_a + len_b, a, b)
      | (len_a, len_b) when len_a <= len_b ->
          let root = nth b 0
          and b = remove b 0
          in Node (root, len_a + len_b, a, b)
      | _ -> failwith "No match"

  let rec replace f tree i =
    match tree with
      | Node (x, sz, a, b) ->
          ( match i - length a with
              | n when n < 0 ->
                  Node (x, sz, replace f a i, b)
              | n when n > 0 ->
                  Node (x, sz, a, replace f b (n - 1))
              | 0 ->
                  Node (f x, sz, a, b)
              | _ -> failwith "No match"
          )
      | _ -> raise Not_found

  let find_first f beg tree =
    let rec aux stack =
      match stack with
        | [] -> -1
        | (i, Node (_, n, _, _))::q when i + n <= beg -> aux q
        | (_, Leaf)::q -> aux q
        | (i, Node (x, 1, Leaf, Leaf))::q ->
            if f x then i else aux q
        | (i, Node (x, n, a, b))::q ->
            let len_a = length a
            in aux ( (i, a) ::
                     (i + len_a, Node (x, 1, Leaf, Leaf)) ::
                     (i + 1 + len_a, b) ::
                     q
            )
    in aux [(0, tree)]

  let rec map f = function
    | Leaf -> Leaf
    | Node (x, n, a, b) ->
        Node (f x, n, map f a, map f b)

  let fold_left f acc tree =
    let rec aux acc = function
      | [] -> acc
      | Leaf::q -> aux acc q
      | (Node (x, 1, Leaf, Leaf))::q -> aux (f acc x) q
      | (Node (x, _, a, b))::q ->
          aux acc (a::(single x)::b::q)
    in aux acc [tree]

  let rec from_list = function
    | [] -> Leaf
    | c::[] -> single c
    | l::c::[] -> Node (c, 2, single l, Leaf)
    | l::c::r::[] -> Node (c, 3, single l, single r)
    | lst -> 
        let len = List.length lst in
        let (l, c::r) = List.cut lst (len / 2)
        in Node (c, len, from_list l, from_list r)

  let to_list t =
    let rec aux acc t =
      let last = length t - 1 in
      if last < 0
        then acc
        else aux ((nth t last)::acc) (remove t last)
    in aux [] t
end
