(* ************************Package More************************************ *)
(* ------------------------------------------------------------------------ *)
(** Generic modules. 
 * This includes standard library enhancements 
 * and some usefull structures and functions. *)

(**{2 Enhanced standard modules }*)

(*Improved List module*)

module List
= struct

  include List

  (** Replaces the nth element of the list by its image by the function. *)
  let replace f lst n =
    let rec aux acc pos = function
      | [] -> lst
      | t::q -> if pos = n
                  then List.rev_append acc ((f t)::q)
                  else aux (t::acc) (pos+1) q
    in aux [] 0 lst

  (** Returns the list without the element.
   * @raise Not_found if a isn't a member of lst. *)
  let remove a lst =
    let rec aux acc = function
      | [] -> raise Not_found
      | t::q -> if a = t
                  then List.rev_append acc q
                  else aux (t::acc) q
    in aux [] lst

  (** [List.cut lst nb] cuts the list [lst] in two lists. If [lst] is long
   * enough, the first returned list will have exactly [n] elements. *)
  let cut lst nb =
    let rec aux acc n l =
      if n >= nb then (rev acc, l)
      else
        match l with
          | [] -> (rev acc, [])
          | t::q -> aux (t::acc) (n+1) q
    in aux [] 0 lst

  (** Returns the list of all (unordered) combinations of the elements of the
   * list. *)
  let combi l =
    let rec first acc = function
      | []   -> acc
      | t::q -> first (second acc t q) q
    and second acc a = function
      | []   -> acc
      | t::q -> second ((a,t)::acc) a q
    in first [] l

  (** Returns the cartesian product of two lists. *)
  let cartesian_product lst_a lst_b =
    List.fold_left
      (fun acc_a v_a ->
        List.fold_left (fun acc_b v_b -> (v_a, v_b)::acc_b) acc_a lst_b)
      []
      lst_a

end

(*Improved String module*)

module String
= struct

  include String

  (** Number of Unicode glyphs from an UTF8 string *)
  let utf8_length str =
    let length = length str in
    let rec aux acc pos =
      if pos >= length then acc else
      match (Char.code str.[pos]) with
        | n when n < 192 -> aux (acc+1) (pos+1)
        | n when n < 224 -> aux (acc+1) (pos+2)
        | n when n < 240 -> aux (acc+1) (pos+3)
        | n when n < 248 -> aux (acc+1) (pos+4)
        | n when n < 252 -> aux (acc+1) (pos+5)
        | _ -> aux (acc+1) (pos+6)
    in aux 0 0

end

(**{2 Structures on [int] }*)

(*Implementation of OrderedType from Module Set*)

module IntOrdered
= struct
  type t = int

  let compare = (-)
end

(*Improved Int Set,using IntOrdered as Implementation for Make Functor, in Set module*)

module IntSet
= struct
  include Set.Make (IntOrdered)

  (** Test if the two sets have a non-empty intersection. *)
  let has_inter a b =
    exists (fun e -> mem e a) b

  (** Build an Int Set based on an int list through List.fold_left. *)
  let from_list (lst: int list) =
    List.fold_left (fun set elt -> add elt set) empty lst

  (** Integers from a to b included *)
  let interval (a:int) (b:int) =
    let rec aux acc p =
      if p > b then acc else aux (add p acc) (p + 1)
    in aux empty a

end

(**{2 Miscelanous Functions }*)

(** [More.swapif b x y] returns [(y,x)] if [b] is [true] and [(x,y)] otherwise.
 *)
let swapif b x y = if b then (y,x) else (x,y)

(** This function operates on injective partial functions from
 * {% $\mathbb\{ %}N{% \}$ %} to {% $\mathbb\{ %}N{% \}$ %} represented
 * as lists of integer pairs. It returns the union of two injective partial
 * functions if this union is still an injective partial function and [None]
 * otherwise. *)
let int_inject_union a b =
  let rec aux acc la lb =
    match (la, lb) with
      | ([], []) ->
          Some acc

      | (_,  []) ->
          Some (List.rev_append acc la)

      | ([], t::q)  ->
          aux [] (t::acc) q

      | ((fa,ta)::qa, (fb,tb)::qb) ->
          let df = fa - fb
          and dt = ta - tb in
          if df * dt = 0 (* one pair (at least) equal*)
            then
              if df + dt = 0 (* both pairs equal *)
                then aux [] (List.rev_append acc la) qb
                else None
          else aux ((fa,ta)::acc) qa lb

  in match a with
      | [] -> Some b
      | _  -> aux [] a b
