(* Package Proof *)
(* ------------- *)
(** Generic operation on proof trees.
 *)

open More

module type THEORIE = 
sig
  type seq
  type rule_un
  type rule_bin

  val seq_equal : seq -> seq -> bool

  val is_axiom : seq -> bool
  val func_un  : rule_un  -> ( seq -> seq )
  val func_bin : rule_bin -> ( seq -> seq -> seq )

  val match_un  : seq -> seq -> rule_un
  val match_bin : seq -> seq -> seq -> rule_bin

  val string_of_seq : seq -> string
  val string_of_rule_un  : rule_un  -> string
  val string_of_rule_bin : rule_bin -> string
end

module type S =
sig
  type seq
  type optree
  type seqtree
  type prooftree

  val fromOp  : optree  -> prooftree
  val fromSeq : seqtree -> prooftree
  val toOp  : prooftree -> optree
  val toSeq : prooftree -> seqtree

  val result : prooftree -> seq
  val verify : prooftree -> bool

  val aff : prooftree -> ColFormat.column
  val printer : Format.formatter -> prooftree -> unit
end

module Make (T : THEORIE)
: (S with type seq       = T.seq
      and type optree    = (T.seq, T.rule_un, T.rule_bin) OneTwoTree.tree
      and type seqtree   = (T.seq, T.seq, T.seq) OneTwoTree.tree
      and type prooftree = (T.seq, T.rule_un, T.rule_bin) OneTwoTree.prooftree)
= struct
  type seq       = T.seq
  type optree    = (T.seq, T.rule_un, T.rule_bin) OneTwoTree.tree
  type seqtree   = (T.seq, T.seq, T.seq) OneTwoTree.tree
  type prooftree = (T.seq, T.rule_un, T.rule_bin) OneTwoTree.prooftree

  let fromOp =
    OneTwoTree.pseudo_map_down
      (fun seq         op -> (op, T.func_un  op seq))
      (fun seq_a seq_b op -> (op, T.func_bin op seq_a seq_b))

  let fromSeq =
    OneTwoTree.pseudo_map_down
      (fun seq_a       seq_c -> (T.match_un  seq_a       seq_c, seq_c))
      (fun seq_a seq_b seq_c -> (T.match_bin seq_a seq_b seq_c, seq_c))

  let toOp t =
    OneTwoTree.map_down (fun x -> x) fst fst t

  let toSeq t =
    OneTwoTree.map_down (fun x -> x) snd snd t

  let result t =
    OneTwoTree.fold_down
      (fun s ->
        if T.is_axiom s then s
          else failwith "Not an axiom")
      (fun s (op, r) ->
        if T.seq_equal (T.func_un op s) r then r
          else failwith "Unary op error")
      (fun a b (op, r) ->
        if T.seq_equal (T.func_bin op a b) r then r
          else failwith "Binary op error")
      t

  let verify t =
    (* TODO : use Exceptions *)
    try (ignore (result t) ; true)
    with _ -> false

  let aff t =
    let append seq op col =
      let str = T.string_of_seq seq in
      let line =
        let str_len = String.utf8_length str
        and op_len  = String.utf8_length op + 1 in
        if op_len > 1 && str_len > op_len
          then
            (String.make (str_len - op_len) '-') ^ " " ^ op
          else
            String.make (String.utf8_length str) '-'
      in ColFormat.append_center (ColFormat.append_center col line) str
    in OneTwoTree.fold_down
      (fun s -> append s "Axiom" ColFormat.new_column)
      (fun col (op, seq) -> append seq (T.string_of_rule_un op) col)
      (fun col_a col_b (op, seq) ->
          append seq (T.string_of_rule_bin op) (ColFormat.join [col_a;col_b]))
      t

  let printer fmt proof =
    ColFormat.printer fmt (aff proof)

end
